/******************************************************************************

   This program is free software; you can redistribute it and/or
   modify it under the terms of the GNU General Public License
   as published by the Free Software Foundation; either version 2
   of the License, or (at your option) any later version.
   
   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.
   
   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

******************************************************************************/

#include "bufferedsettings.hpp"

BufferedSettings::BufferedSettings(const std::string& filename) : Settings(filename) {
    for (DOMNode *childNode = parser->getDocument()->getDocumentElement()->getFirstChild(); childNode != NULL; childNode = childNode->getNextSibling()) {        
        buildBuffer(childNode, "");       
    }
    // Clear memory pool as we will not use it anymore
    parser->resetDocumentPool();
}

BufferedSettings::~BufferedSettings() {

}

void BufferedSettings::insert(const std::string& key, const std::string& value) {
    MapStrPStrInt::const_iterator first = settings.lower_bound(key);
    int max = 0;
    if (first != settings.end()) {
        MapStrPStrInt::const_iterator last = settings.upper_bound(key);
        for (MapStrPStrInt::const_iterator iter = first; iter != last; ++iter) {
            if (iter->second.second > max) {
                max = iter->second.second;
            }
        }
        ++max;
    }
    settings.insert(std::make_pair(key, std::make_pair(value, max)));     
}

void BufferedSettings::buildBuffer(DOMNode* currentNode, std::string path) {
    if (currentNode->getNodeType() == DOMNode::TEXT_NODE) {
        insert(path.substr(1), XMLtr(currentNode->getNodeValue()));
    }
    for (DOMNode *childNode = currentNode->getFirstChild();
         childNode != NULL;
         childNode = childNode->getNextSibling()) {
        buildBuffer(childNode, path + "/" + XMLtr(currentNode->getNodeName()));
    }
}

std::set<std::string> BufferedSettings::allParameters(const std::string& path, const std::string& param) const {
    std::set<std::string> ret;
    MapStrPStrInt::const_iterator first = settings.lower_bound(path + "/" + param);
    MapStrPStrInt::const_iterator last = settings.upper_bound(path + "/" + param);
    for (MapStrPStrInt::const_iterator iter = first; iter != last; ++iter) {
        ret.insert(iter->second.first);
    }
    return ret;
}

Settings::SubParameters BufferedSettings::allSubParameters(const std::string& base, const std::list<std::string>& params) const {
    SubParameters ret;
    MapStrPStrInt::const_iterator first = settings.lower_bound(base + "/" + params.front());
    MapStrPStrInt::const_iterator last = settings.upper_bound(base + "/" + params.front());
    for (MapStrPStrInt::const_iterator iter = first; iter != last; ++iter) {
        std::list<std::string> tmpret;
        int currentIndex = iter->second.second;
        std::list<std::string>::const_iterator param_iter = params.begin();
        tmpret.push_back(iter->second.first);
        ++param_iter; // ignore first param, already used...
        for (/*empty*/; param_iter != params.end(); ++param_iter) {
            MapStrPStrInt::const_iterator value_first = settings.lower_bound(base + "/" + *param_iter);
            MapStrPStrInt::const_iterator value_last = settings.upper_bound(base + "/" + *param_iter);
            bool found = false;
            for (MapStrPStrInt::const_iterator value_iter = value_first; value_iter != value_last; ++value_iter) {
                if (value_iter->second.second == currentIndex) {
                    found = true;
                    tmpret.push_back(value_iter->second.first);
                }
            }
            if (!found) {
                std::cerr << "*** ### *** Fatalt fel" << std::endl;
                throw NotFoundException("Parameter '" + *param_iter + "' not found.");
            }
        }
        ret.insert(tmpret);
    }
    return ret;
}

std::string BufferedSettings::parameterValue(const std::string& path) const throw(ParamNotFoundException) {
    MapStrPStrInt::const_iterator iter = settings.find(path);
    if (iter == settings.end()) {
        throw ParamNotFoundException();
    }
    return iter->second.first;
}

void BufferedSettings::addParameter(const std::string& path, const std::string& value) {
    settings.erase(path);
    insert(path, value);
}

bool BufferedSettings::insertCommandLineParameters(int argc, char* argv[]) {
    for (int i = 1; argv[i]; ++i) {
        char *arg = argv[i];
        if (*arg == '-') {
            switch (*++arg) {
            case 'd': 
                addParameter("damerau/active", "true");
                if (*++arg != '\0')
                    addParameter("damerau/probability", arg);
                break;
            case 's': 
                addParameter("splitCompound/active", "true");
                if (*++arg != '\0')
                    addParameter("splitCompound/probability", arg);
                break;
            case 'e': 
                addParameter("soundError/active", "true");
                if (*++arg != '\0')
                    addParameter("soundError/probability", arg);
                break;
            case 'i': 
                if (*++arg == '\0') return false; 
                addParameter("input/filename", arg); break;
            case 'o': 
                if (*++arg == '\0') return false; 
                addParameter("output/filename", arg); break;
            case 'x': 
                if (*++arg == '\0') return false; 
                addParameter("errorWriter/filename", arg); break;
            case 'n':
                if (*++arg == '\0') return false;
                addParameter("options/seed", arg); break;
            case 'c':
                if (*++arg == '\0') return false;
                addParameter("options/configDir", arg); break;
            case 'w':
                if (*++arg == '\0') return false;
                addParameter("wordlist/create/filename", arg); break;
            case '-':
                if (*++arg == '\0') return false;
                std::string param = arg;
                std::string::size_type pos = param.find_first_of('=');
                if (pos == std::string::npos) {
                    return false;
                }
                addParameter(param.substr(0, pos), param.substr(pos+1));
            }
        }
        else if (*arg == '+') {
            switch (*++arg) {
            case 'd': addParameter("damerau/active", "false"); break;
            case 's': addParameter("splitCompound/active", "false"); break;
            case 'e': addParameter("soundError/active", "false"); break;
            default: return false;
            }
        }
    }
    try {
        boost::lexical_cast<int>(parameterValue("options/seed"));
    }
    catch (const ParamNotFoundException e) {
        addParameter("options/seed", boost::lexical_cast<std::string>(std::time(0)));
    }
    catch (boost::bad_lexical_cast &) {
        addParameter("options/seed", boost::lexical_cast<std::string>(std::time(0)));
    }
    return true;
}
