/******************************************************************************

   This program is free software; you can redistribute it and/or
   modify it under the terms of the GNU General Public License
   as published by the Free Software Foundation; either version 2
   of the License, or (at your option) any later version.
   
   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.
   
   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

******************************************************************************/

#include "errorintroducer.hpp"

ErrorIntroducer::ErrorIntroducer(Wordlist &wl, Settings &se, ErrorWriter &ew, InputReader &ir, const std::string &filename) 
     : wordlist(wl)
     , settings(se)
     , errorwriter(ew)
     , inputreader(ir)
     , escapeChar(settings.getChar("options/escapeChar", '\\'))
     , wordChar(settings.getChar("options/wordChar", '\'')) 
     , allowExistingWords(settings.getBool("errorIntroducer/allowExistingWords", true))
     , forceAllowWords(settings.getBool("errorIntroducer/forceAllowWords", false))
     , allowTagChange(settings.getBool("errorIntroducer/allowTagChange", true))
     , forceAllowTag(settings.getBool("errorIntroducer/forceAllowTag", false)) 
     , reportErrorIntroducer(settings.getString("errorIntroducer/reportName", "introduced")) {
     spaceChar = settings.getChar("options/spaceChar", ' ');
     if (filename != "") {
	  readFromFile(filename);
     }
}

void ErrorIntroducer::readFromFile(const std::string &filename) {
     XercesDOMParser *parser = new XercesDOMParser();
     ErrorHandler *errHandler = new MyErrorHandler();

     parser->setCreateCommentNodes(false);

     // parser->setDoValidation(true);
     xercesc::AbstractDOMParser::ValSchemes    valScheme = xercesc::AbstractDOMParser::Val_Auto;
     parser->setValidationScheme(valScheme);
     
     parser->setDoNamespaces(true);
     parser->setDoSchema(true);
     parser->setIncludeIgnorableWhitespace(false);
     parser->setValidationScheme(XercesDOMParser::Val_Always);
     parser->setErrorHandler(errHandler);

     try {
	  parser->parse(filename.c_str());
     }
     catch (const XMLException& toCatch) {
	  char* message = XMLtr(toCatch.getMessage());
	  std::cerr << "Exception message is: \n"
	       << message << "\n";
	  delete [] message;
	  XMLPlatformUtils::Terminate();
	  std::exit(1);
     }
     catch (const DOMException& toCatch) {
	  std::cerr << "Exception message is: \n"
		    << toCatch.code << "\n";
	  XMLPlatformUtils::Terminate();
	  std::exit(1);
     }
     catch (...) {
	  std::cerr << "Unexpected XML parse exception\n"
		    << "Probably malformed error file.\n";
	  XMLPlatformUtils::Terminate();
	  std::exit(1);
     }
     
     DOMElement *root = parser->getDocument()->getDocumentElement();
     if (XMLcmp(root->getTagName(), XMLtr("errorIntroducer")) != 0) {
	  std::cerr << "This is not an errorIntroducer file!\n";
     }

     Separator sep(spaceChar.c_str());

     for (DOMNode *node = root->getFirstChild(); 
	  node != NULL; node = node->getNextSibling()) {
	  try {
	       if (node->getNodeType() == DOMNode::ELEMENT_NODE
		   && (XMLcmp(node->getNodeName(), XMLtr("rule")) == 0)) {
		    std::string match;
		    std::string to;
		    std::string order;
		    for (DOMNode *child = node->getFirstChild();
			 child != NULL; child = child->getNextSibling()) {
			 if (child->getNodeType() == DOMNode::ELEMENT_NODE) {
			      if (XMLcmp(child->getNodeName(), XMLtr("match")) == 0) {
				   for (DOMNode *dn = child->getFirstChild(); 
					dn != NULL; dn = dn->getNextSibling()) {
					if (dn->getNodeType() == DOMNode::TEXT_NODE) {
					     match = XMLtr(dn->getNodeValue());
					}	  
				   }
			      }
			      if (XMLcmp(child->getNodeName(), XMLtr("to")) == 0) {
				   for (DOMNode *dn = child->getFirstChild(); 
					dn != NULL; dn = dn->getNextSibling()) {
					if (dn->getNodeType() == DOMNode::TEXT_NODE) {
					     to = XMLtr(dn->getNodeValue());
					}	  
				   }
			      }
			      if (XMLcmp(child->getNodeName(), XMLtr("order")) == 0) {
				   for (DOMNode *dn = child->getFirstChild(); 
					dn != NULL; dn = dn->getNextSibling()) {
					if (dn->getNodeType() == DOMNode::TEXT_NODE) {
					     order = XMLtr(dn->getNodeValue());
					}	  
				   }
			      }
			 }			 
		    }
		    if (match.at(0) >= 'a' && match.at(0) <= 'z') {
			 sortedRules.insert(boost::make_tuple(match, to, order));
		    }
		    else {
			 unsortedRules.push_back(boost::make_tuple(match, to, order));
		    }
	       }
	  }
	  catch (std::out_of_range &) {
	       std::cerr << "Warning: Out of range\n";
	  }
	  catch (...) {
	       std::cerr << "Error: Unexpected exception\n";
	  }
     }
     delete parser;
     delete errHandler;
}

void ErrorIntroducer::printList(const TupleList &listTuple) const {
     std::cerr << "--- begin listTuple ---\n";
     for (TupleList::const_iterator citer = listTuple.begin(); citer != listTuple.end(); ++citer) {
	  std::cerr << boost::get<0>(*citer).word() << " " << boost::get<0>(*citer).tag() << " " << boost::get<0>(*citer).row();
	  std::cerr << "\t|\t" << boost::get<1>(*citer).word() << " " << boost::get<1>(*citer).tag() << " " << boost::get<1>(*citer).row();
	  std::cerr << "\t|\t" << boost::get<2>(*citer) << "\n";
     }
     std::cerr << "--- end listTuple ---\n";
}

void ErrorIntroducer::printVector(const TupleVector &resultTuple) const {
     std::cerr << "--- begin vectorTuple ---\n";
     for (TupleVector::const_iterator citer = resultTuple.begin(); citer != resultTuple.end(); ++citer) {
	  std::cerr << boost::get<0>(*citer).word() << " " << boost::get<0>(*citer).tag() << " " << boost::get<0>(*citer).row();
	  std::cerr << "\t|\t" << boost::get<1>(*citer).word() << " " << boost::get<1>(*citer).tag() << " " << boost::get<1>(*citer).row();
	  std::cerr << "\t|\t" << boost::get<2>(*citer) << "\n";
     }
     std::cerr << "--- end vectorTuple ---\n";
}

void ErrorIntroducer::newOrder(const std::string& match, const std::string& to, const std::string& order, std::list<WordTag>& listwt) const {
     TupleVector vectorTuple;
     TupleVector resultTuple;
     TupleList listTuple;

     Separator sep(spaceChar.c_str());
     Tokenizer matchToken(match, sep);
     Tokenizer toToken(to, sep);
     Tokenizer::const_iterator matchIter = matchToken.begin();
     Tokenizer::const_iterator toIter = toToken.begin();

     MinEdit::OperationList operations;
     int num_match = 0;

     { // just a new scope
	  std::list<WordTag>::const_iterator biter = listwt.begin();
	  for (Tokenizer::const_iterator citer = matchToken.begin();
	       citer != matchToken.end(); ++citer) {
	       listTuple.push_back(boost::make_tuple(*biter, *biter, false));
	       vectorTuple.push_back(boost::make_tuple(*biter, *biter, false));
	       ++biter;
	       ++num_match;
	  }
     }

     // Perform word and wordtag changes
     TupleList::iterator listPos = listTuple.begin();
     TupleVector::iterator vectorPos = vectorTuple.begin();
     for (/*nothing*/; matchIter != matchToken.end() && toIter != toToken.end(); 
	   ++matchIter, ++toIter) {
	  if ((*toIter)[0] == wordChar) {
	       // This is a literal word
	       std::string oldword = boost::get<1>(*vectorPos).word();
	       std::string newword = (*toIter).substr(1, (*toIter).length() - 2);
	       if (newword != oldword) {
		    int oldtag = boost::get<1>(*vectorPos).tag();
		    int newtag = wordlist.getTag(newword, oldtag);
		    if (wordlist.lookup(newword) && !allowExistingWords) {
			 throw OperationNotPermitted("'" + newword + "' exist, and new words are not allowed to exist.");
		    }
		    if (!wordlist.lookup(newword) && allowExistingWords && forceAllowWords) {
			 throw OperationNotPermitted("'" + newword + "' doesn't exist, and new words have to exist.");
		    }
		    if (oldtag != newtag && !allowTagChange) {
			 throw OperationNotPermitted("Tag changed, and that is not allowed.");
		    }
		    if (oldtag == newtag && allowTagChange && forceAllowTag) {
			 throw OperationNotPermitted("No tag change, and tags must change.");
		    }
		    boost::get<1>(*vectorPos).set(newword, newtag);
		    boost::get<2>(*vectorPos) = true;
		    boost::get<1>(*vectorPos).addDesc(reportErrorIntroducer);
		    boost::get<1>(*listPos).set(newword, newtag);
		    boost::get<2>(*listPos) = true;
		    boost::get<1>(*listPos).addDesc(reportErrorIntroducer);
	       }
	  }
	  else {
	       // This is a word tag
	       if (*matchIter != *toIter) {
		    std::string oldword = boost::get<1>(*vectorPos).word();
		    int oldtag = boost::get<1>(*vectorPos).tag();
		    int newtag = wordlist.untag(*toIter);
		    std::string lemma = wordlist.findLemma(oldword, oldtag);
		    std::string newword = wordlist.findWord(lemma, newtag);
		    if (oldtag != newtag) {
			 if (wordlist.lookup(newword) && !allowExistingWords) {
			      throw OperationNotPermitted("'" + newword + "' exist, and new words are not allowed to exist.");
			 }
			 if (!wordlist.lookup(newword) && allowExistingWords && forceAllowWords) {
			      throw OperationNotPermitted("'" + newword + "' doesn't exist, and new words have to exist.");
			 }
			 if (!allowTagChange) {
			      throw OperationNotPermitted("Tag changed, and that is not allowed.");
			 }
			 boost::get<1>(*vectorPos).set(newword, newtag);
			 boost::get<2>(*vectorPos) = true;
			 boost::get<1>(*vectorPos).addDesc(reportErrorIntroducer);
			 boost::get<1>(*listPos).set(newword, newtag);
			 boost::get<2>(*listPos) = true;
			 boost::get<1>(*listPos).addDesc(reportErrorIntroducer);
		    }
	       }
	  }
	  ++vectorPos;
	  ++listPos;
     }

     // Fill vectorTuple
     for (/*nothing*/; toIter != toToken.end(); ++toIter) {
	  if ((*toIter)[0] == wordChar) {
	       // This is a literal word
	       std::string newword = (*toIter).substr(1, (*toIter).length() - 2);
	       vectorTuple.push_back(boost::make_tuple(WordTag(newword, wordlist.getTag(newword, wordlist.untag(wordlist.unknownTag())), 0, "", wordlist.untag(wordlist.unknownTag()), 0), WordTag(newword, wordlist.getTag(newword, wordlist.untag(wordlist.unknownTag())), 0, "", wordlist.untag(wordlist.unknownTag()), 0), true));
	  }
	  else {
	       std::cerr << "Fatal error!\n";
	  }
     }

     try {
	  Separator sep(spaceChar.c_str());
	  Tokenizer matchToken(match, sep);
	  Tokenizer toToken(to, sep);
	  std::string from;
	  int i = 0;

	  // Create the string to be compared with the 'order' string
	  for (Tokenizer::const_iterator citer = matchToken.begin(); citer != matchToken.end(); ++citer) {
	       ++i;
	       from += boost::lexical_cast<std::string>(i);
	  }

	  // Compute the final order of the tags
	  // FIXME use index?
	  for (std::string::const_iterator citer = order.begin(); 
	       citer != order.end(); ++citer) {
	       int index = boost::lexical_cast<int>(*citer);
	       TupleVector::const_iterator resIter = vectorTuple.begin();
	       for (int i = 1; i < index; ++i) {
		    ++resIter;
	       }
	       resultTuple.push_back(*resIter);
	  }

	  // Calculate shortest list of operations (move, insert delete)
	  operations = MinEdit::distance2(from, order);

     }
     catch (const boost::bad_lexical_cast &) {
	  std::cerr << "Warning: <order> element must only contain numbers\n";
     }
     catch (...) {
	  std::cerr << "Fatal error.\n";
     }

     // Update row indication
     unsigned int row_number = listwt.front().row();
     for (vectorPos = resultTuple.begin(); vectorPos != resultTuple.end(); ++vectorPos) {
	  boost::get<1>(*vectorPos).setRow(row_number);
	  ++row_number;
     }

     // Register the changes
     for (MinEdit::OperationList::const_iterator citer = operations.begin();
	  citer != operations.end(); ++citer) {
	  char first = boost::get<0>(*citer);
	  int second = boost::get<1>(*citer);
	  int third = boost::get<2>(*citer);
	  // Move
	  if (first == 'm') {
	       TupleList::iterator liter = listTuple.begin();
	       for (int j = 1; j != second && liter != listTuple.end(); ++liter) {
		    ++j;
	       }
	       TupleList::iterator bliter = listTuple.begin();
	       for (int k = 1; k != third && bliter != listTuple.end(); ++bliter) {
		    ++k;
	       }
	       boost::get<2>(*liter) = false;
	       std::string oldword = boost::get<0>(*liter).word();
	       int oldtag = boost::get<0>(*liter).tag();
	       std::string newword = boost::get<1>(*liter).word();
	       newword = wordlist.preserveCase(oldword, newword);
	       int newtag = boost::get<1>(*liter).tag();
	       int orig = boost::get<0>(*liter).orig();
	       errorwriter.insertIntroduced("move", orig, listwt.front().row() + second - 1, listwt.front().row() + third - 1, oldword, oldtag, newword, newtag);
	       listTuple.splice(bliter, listTuple, liter);
	  }
	  // Delete
	  else if (first == 'd') {
	       TupleList::iterator liter = listTuple.begin();
	       for (int j = 1; j != second && liter != listTuple.end(); ++liter) {
		    ++j;
	       }
	       boost::get<2>(*liter) = false;
	       std::string oldword = boost::get<0>(*liter).word();
	       int oldtag = boost::get<0>(*liter).tag();
	       std::string newword = "";
	       int newtag = wordlist.untag(wordlist.unknownTag());
	       int orig = boost::get<0>(*liter).orig();
	       errorwriter.insertIntroduced("delete", orig, listwt.front().row() + second - 1, -1, oldword, oldtag, newword, newtag);
	       listTuple.erase(liter);
	  }
	  // Insert
	  else if (first == 'i') {
	       TupleList::iterator liter = listTuple.begin();
	       for (int j = 1; j != second && liter != listTuple.end(); ++liter) {
		    ++j;
	       }
	       boost::get<2>(*liter) = false;
	       std::string origWord = boost::get<0>(*liter).word();
	       int origTag = boost::get<0>(*liter).tag();
	       std::string oldword = "";
	       int oldtag = wordlist.untag(wordlist.unknownTag());
	       std::string newword = boost::get<1>(resultTuple[third - 1]).word();
	       int newtag = boost::get<1>(resultTuple[third - 1]).tag();
	       errorwriter.insertIntroduced("insert", -1, -1, listwt.front().row() + second - 1, oldword, oldtag, newword, newtag);
	       listTuple.insert(liter, boost::make_tuple(WordTag(newword, newtag, 0, origWord, origTag, 0), WordTag(newword, newtag, 0, origWord, origTag, 0), false));
	  }
	  else {
	       std::cerr << "Fatal error.\n";
	  }
     }

     // Make sure all changes are registered
     {
	  TupleVector::const_iterator riter = resultTuple.begin();
	  for (TupleList::iterator iter = listTuple.begin(); 
	       iter != listTuple.end() && riter != resultTuple.end(); ++iter, ++riter) {
	       if (boost::get<2>(*iter) == true) {	       
		    std::string oldword = boost::get<0>(*riter).word();
		    int oldtag = boost::get<0>(*riter).tag();
		    int oldrow = boost::get<0>(*riter).row();
		    std::string newword = boost::get<1>(*riter).word();
		    newword = wordlist.preserveCase(oldword, newword);
		    int newtag = boost::get<1>(*riter).tag();
		    int newrow = boost::get<1>(*riter).row();
		    int orig = boost::get<0>(*riter).orig();
		    errorwriter.insertIntroduced("change", orig, oldrow, newrow, oldword, oldtag, newword, newtag);
	       }
	  }
     }

     //printList(listTuple);
     //printVector(resultTuple);

     // Remove the old elements
     for (int i = 0; i < num_match; ++i) {
	  listwt.pop_front();
     }

     // Calculate correct row number
     for (std::list<WordTag>::iterator iter = listwt.begin();
	  iter != listwt.end(); ++iter) {
	  iter->setRow(iter->row() + resultTuple.size() - num_match);
     }
     inputreader.changeReadOffset(resultTuple.size() - num_match);

     // Insert new elements
     for (int i = resultTuple.size() - 1; i >= 0; --i) {
	  WordTag &wtold = boost::get<0>(resultTuple[i]);
	  WordTag &wtinsert = boost::get<1>(resultTuple[i]);
	  wtinsert.setWord(wordlist.preserveCase(wtold.word(), wtinsert.word()));
	  listwt.push_front(wtinsert);
     }
}

void ErrorIntroducer::printRules() const {
     std::cout << "Sorted rules:\n";
     for (RuleSet::const_iterator iter = sortedRules.begin();
	  iter != sortedRules.end(); ++iter) {
	  std::cout << boost::tuples::set_open('[') 
		    << boost::tuples::set_close(']') 
		    << boost::tuples::set_delimiter('#') 
		    << *iter << "\n";
     }
     std::cout << "Unsorted rules:\n";
     for (unsigned int i = 0; i < unsortedRules.size(); ++i) {
	  std::cout << boost::tuples::set_open('[') 
		    << boost::tuples::set_close(']') 
		    << boost::tuples::set_delimiter('#') 
		    << unsortedRules[i] << "\n";
     }
}

bool ErrorIntroducer::applyHelper(const std::string& match, const std::string& to, const std::string& order, std::list<WordTag>& listwt) const {
     boost::regex expression;
     boost::match_results<std::string::const_iterator> result;
     Separator sep(spaceChar.c_str());
     std::string toMatch;
     try {
	  // Parse the <match>-tag and format the input
	  // string in the corresponding way
	  Tokenizer token(match, sep);
	  std::list<WordTag>::const_iterator listPos = listwt.begin();
	  for (Tokenizer::const_iterator itok = token.begin();
	       itok != token.end(); ++itok) {
	       if (listPos->blank()) { 	// Skip "blank" lines
		    return false;
	       }
	       if ((*itok)[0] == wordChar) { // literal word
		    toMatch += wordChar + listPos->word() + wordChar + spaceChar;
	       }
	       else { // word tag
		    toMatch += wordlist.tag(listPos->tag()) + spaceChar;
	       }
	       ++listPos;
	       if (listPos == listwt.end()) {
		    throw EndOfInputException();
	       }
	  }
	  // Match the input string with the <match>-tag
	  toMatch = toMatch.substr(0, toMatch.length() - 1);
	  expression.set_expression(match, boost::regbase::normal);
	  if (regex_match(toMatch, result, expression)) {
	       // Transform according to the <to>-tag
	       std::string newMatch = replaceMatchedExpressions(to, result); 
	       //std::cerr << "New string is: " << newMatch << "\n";

	       try {
		    newOrder(toMatch, newMatch, order, listwt);
	       }
	       catch (const NotFoundException &e) {
		    std::cerr << "Warning: " << e.what() << ". No change done.\n";
		    return false;
	       }
	       catch (const OperationNotPermitted &e) {
		    std::cerr << "Warning: " << e.what() << "\n";
		    return false;
	       }
	       catch (...) {
		    std::cerr << "Unable to convert ord reorder.\n";
		    return false;
	       }
	       return true;
	  }
     }
     catch (EndOfInputException &) {
	  //std::cerr << "End of input. No match even possible for this rule\n";
     }
     return false;
}

bool ErrorIntroducer::apply(std::list<WordTag>& listwt) const {
     Separator sep(spaceChar.c_str());
     try {
	  std::string currentTag = wordlist.tag(listwt.front().tag());
	  // Test sorted rules first
	  if (currentTag[0] >= 'a' && currentTag[0] <= 'z') {
	       std::pair<RuleSet::const_iterator, RuleSet::const_iterator> range = sortedRules.equal_range(boost::make_tuple(currentTag, "dummy", "dummy"));
	       for (RuleSet::iterator iter = range.first; iter != range.second; ++iter) {
		    bool result;
		    result = applyHelper(iter->get<0>(), iter->get<1>(), iter->get<2>(), listwt);
		    if (result) {
			 return true;
		    }
	       }
	  }
	  // unsorted rules: always
	  for (RuleVector::const_iterator iter = unsortedRules.begin(); 
	       iter != unsortedRules.end(); ++iter) {
	       bool result;
	       result = applyHelper(iter->get<0>(), iter->get<1>(), iter->get<2>(), listwt);
	       if (result) {
		    return true;
	       }
	  }
     }
     catch (TagNotFoundException &) {
	  std::cerr << "Warning: Tag not found.";
     }
     catch (...) {
	  std::cerr << "ErrorIntroducer::apply unexpected exception\n";
     }
     return false;
}

std::string ErrorIntroducer::replaceMatchedExpressions(const std::string& exp, const boost::match_results<std::string::const_iterator> &result) const {
     std::string ret = exp;
     std::string::size_type index = ret.find(escapeChar, 0);
     try {
	  while (index != std::string::npos) {
	       unsigned int param = boost::lexical_cast<unsigned int>(ret.at(index + 1));
	       if (param >= 1 && param <= result.size() - 1) {
		    ret.replace(index, 2, result[param]);
	       }
	       else {
		    throw std::out_of_range("index in <to> element to big");
	       }
	       index = ret.find(escapeChar, index);
	  }
     }
     catch (const boost::bad_lexical_cast &) {
	  std::cerr << "Error: " << escapeChar << " must be followed by a number\n";
     }
     catch (const std::out_of_range &e) {
	  std::cerr << "Error: out of range: " << e.what() << "\n";
     }
     return ret;
}
