/******************************************************************************

   This program is free software; you can redistribute it and/or
   modify it under the terms of the GNU General Public License
   as published by the Free Software Foundation; either version 2
   of the License, or (at your option) any later version.
   
   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.
   
   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

******************************************************************************/

#include "wordlist.hpp"

extern "C" {
#include <zlib.h>
}

Wordlist::Wordlist(Settings &se)
     : settings(se), wltlong(0), wtcount(0), useNext(true)
     , unknowntag(settings.getString("options/unknownTag", "unknownTag"))
     , unknownlemma(settings.getString("options/unknownLemma", "unknownLemma")) {
     std::string configDir = settings.getString("options/configDir", "");
     std::string wordfile = configDir + settings.getString("wordlist/wordfile", "outfile.gz");
     std::string tagfile = configDir + settings.getString("wordlist/tagfile", "tagfile");
     gzFile gzfile = gzopen(wordfile.c_str(), "rb");
     if (!gzfile) {
	  std::cerr << "Error: Can't open word list '" << wordfile << "' Aborting...\n";
	  std::exit(1);
     }

     std::ifstream tfile(tagfile.c_str());
     if (!tfile) {
	  std::cerr << "Error: Can't open tag list '" << tagfile << "' Aborting...\n";
	  std::exit(1);
     }

     // Read wordlist
     gzread(gzfile, &wtcount, sizeof(wtcount));
     std::cout << "Reading " << wtcount << " sections...";
     wltlong = new Corpus::WordToken[wtcount];
     lwtlong = new Corpus::WordToken[wtcount];
     if (gzread(gzfile, wltlong, wtcount * sizeof(Corpus::WordToken)) != (int)(wtcount * sizeof(Corpus::WordToken))) {
	  std::cerr << "Error reading file '" << wordfile << "'\n";
	  std::exit(1);
     }
     if (gzread(gzfile, lwtlong, wtcount * sizeof(Corpus::WordToken)) != (int)(wtcount * sizeof(Corpus::WordToken))) {
	  std::cerr << "Error reading file '" << wordfile << "'\n";
	  std::exit(1);
     }
     std::cout << "done\n";

     // Read tags
     std::cout << "Reading tagfile " << tagfile << "...";
     std::string line;
     while (std::getline(tfile, line)) {
	  std::istringstream str(line);
	  int num;
	  std::string tag;
	  str >> num >> tag;
	  tags.insert(std::make_pair(num, tag));
	  untags.insert(std::make_pair(tag, num));
     }
     // Add 'unknown' tag
     MapIntStr::const_iterator tagiter = tags.end();
     --tagiter;
     tags.insert(std::make_pair(tagiter->first + 1, unknowntag));
     untags.insert(std::make_pair(unknowntag, tagiter->first + 1));
     
     gzclose(gzfile);
     std::cout << "done\n";
}

Wordlist::~Wordlist() {
     if (wltlong) {
	  delete [] wltlong;
     }
}

Wordlist::const_iterator Wordlist::begin() {
     return WordlistIterator(*this, wltlong);
}

Wordlist::const_iterator Wordlist::end() {
     return WordlistIterator(*this, (Corpus::WordToken *)(wltlong + wtcount));
}

Wordlist::const_iterator Wordlist::lemma_begin() {
     return WordlistIterator(*this, lwtlong);
}

Wordlist::const_iterator Wordlist::lemma_end() {
     return WordlistIterator(*this, (Corpus::WordToken *)(lwtlong + wtcount));
}

int Wordlist::startOfWord(int pos, const Corpus::WordToken wlt[]) const {
     while (true) {
	  if (wlt[pos].data() < 0) {
	       return pos;
	  }
	  else if (wlt[pos].data() > 0) {
	       if (pos == 0) {
		    return pos;
	       }
	       else if (wlt[pos - 1].data() < 0) {
		    return pos - 1;
	       }
	       else if (wlt[pos - 1].data() >= 0) {
		    return pos;
	       }
	  }
	  else {
	       pos--;
	  }
     }
     return 0; // dummy
}

int Wordlist::startOfNextWord(int pos, const Corpus::WordToken wlt[]) const {
     while (true) {
	  if (wlt[pos].data() < 0) {
	       return pos;
	  }
	  else if (wlt[pos].data() > 0) {
	       if (pos == 0) {
		    return pos;
	       }
	       else if (wlt[pos - 1].data() < 0) {
		    pos++;
	       }
	       else if (wlt[pos - 1].data() >= 0) {
		    return pos;
	       }
	  }
	  else {
	       pos++;
	  }
     }
     return 0; // dummy
}

WordlistIterator Wordlist::findFirstWord(const std::string& word) {
     std::string lookfor = word;
     Utils::lowercase(lookfor);
     const Corpus::WordToken *ret = findFirst(lookfor, wltlong);
     if (ret == NULL) {
	  return end();
     }
     return WordlistIterator(*this, ret);
}

WordlistIterator Wordlist::findFirstLemma(const std::string& word) {
     std::string lookfor = word;
     Utils::lowercase(lookfor);
     const Corpus::WordToken *ret = findFirst(lookfor, lwtlong);
     if (ret == NULL) {
	  return lemma_end();
     }
     return WordlistIterator(*this, ret);
}

const Corpus::WordToken* Wordlist::findFirst(const std::string& word, 
					     const Corpus::WordToken wtlong[]) {
     int min = 0;
     int pos;
     int max = startOfWord(wtcount - 1, wtlong);
     // Check if max and min are equal or if their words are adjacent
     while (max != min && (nextToken(min, wtlong) != max)) {
	  pos = (max - min) / 2 + min;
	  pos = startOfSomeWord(pos, wtlong);
	  if (word < wtlong[pos].word()) {
	       max = pos;
	  }
	  else if (word > wtlong[pos].word()) {
	       min = pos;
	  }
	  else {
	       max = pos;
	  }
     }
     if (word == wtlong[min].word()) {
	  return &wtlong[min];
     }
     else if (word == wtlong[max].word()) {
	  return &wtlong[max];
     }
     return NULL;
}

WordlistIterator Wordlist::findLastWord(const std::string& word) {
     std::string lookfor = word;
     Utils::lowercase(lookfor);
     const Corpus::WordToken *ret = findLast(lookfor, wltlong);
     if (ret == NULL) {
	  return end();
     }
     return WordlistIterator(*this, (const Corpus::WordToken *)(ret + ret->size() + 1));
}

WordlistIterator Wordlist::findLastLemma(const std::string& word) {
     std::string lookfor = word;
     Utils::lowercase(lookfor);
     const Corpus::WordToken *ret = findLast(lookfor, lwtlong);
     if (ret == NULL) {
	  return lemma_end();
     }
     return WordlistIterator(*this, (const Corpus::WordToken *)(ret + ret->size() + 1));
}

const Corpus::WordToken* Wordlist::findLast(const std::string& word, 
					    const Corpus::WordToken wtlong[]) {
     int min = 0;
     int pos;
     int max = startOfWord(wtcount - 1, wtlong);
     // Check if max and min are equal or if their words are adjacent
     while (max != min && (nextToken(min, wtlong) != max)) {
	  pos = (max - min) / 2 + min;
	  pos = startOfSomeWord(pos, wtlong);
	  if (word < wtlong[pos].word()) {
	       max = pos;
	  }
	  else if (word > wtlong[pos].word()) {
	       min = pos;
	  }
	  else {
	       min = pos;
	  }
     }
     if (word == wtlong[max].word()) {
	  return &wtlong[max];
     }
     else if (word == wtlong[min].word()) {
	  return &wtlong[min];
     }
     return NULL;
}

const std::string Wordlist::findLemma(const std::string& word, int tag) {
     std::string lookfor = word;
     Utils::lowercase(lookfor);
     Wordlist::const_iterator iter_last = findLastWord(lookfor);
     for (Wordlist::const_iterator iter = findFirstWord(lookfor);
	  iter != iter_last; ++iter) {
	  if (iter->tag() == tag) {
	       return iter->lemma();
	  }
     }
     throw NotFoundException("Lemma for " + word + " not found");
}

const std::string Wordlist::findWord(const std::string& lemma, int tag) {
     std::string lookfor = lemma;
     Utils::lowercase(lookfor);
     Wordlist::const_iterator iter_last = findLastLemma(lookfor);
     for (Wordlist::const_iterator iter = findFirstLemma(lookfor);
	  iter != iter_last; ++iter) {
	  if (iter->tag() == tag) {
	       return iter->lemma();
	  }
     }
     throw NotFoundException("Word for " + lemma + " not found");
}

bool Wordlist::lookup(const std::string& word) {
     std::string lookfor = word;
     Utils::lowercase(lookfor);
     int min = 0;
     int pos;
     int max = startOfWord(wtcount - 1, wltlong);
     while (max != min && (nextToken(min, wltlong) != max)) {
	  pos = (max - min) / 2 + min;
	  pos = startOfSomeWord(pos, wltlong);
	  if (lookfor < wltlong[pos].word()) {
	       max = pos;
	  }
	  else if (lookfor > wltlong[pos].word()) {
	       min = pos;
	  }
	  else {
	       return true;
	  }
     }
     if (lookfor == wltlong[min].word()) {
	  return true;
     }
     else if (lookfor == wltlong[max].word()) {
	  return true;
     }
     return false;
}


inline int Wordlist::nextToken(const int token, const Corpus::WordToken* wtlong) const {
     return token + wtlong[token].size() + 1;
}

const Corpus::WordToken* Wordlist::nextToken(const Corpus::WordToken* token) const {
     return (const Corpus::WordToken *)(token + token->size() + 1);
}

const std::string& Wordlist::tag(int id) const {
     MapIntStr::const_iterator iter = tags.find(id);
     if (iter != tags.end()) {
	  return iter->second;
     }
     throw TagNotFoundException("Tag not found.");
}

int Wordlist::untag(std::string val) {
     Utils::lowercase(val);
     MapStrInt::const_iterator iter = untags.find(val);
     if (iter == untags.end()) {
	  std::cerr << "Note: '" << val << "' is a new tag.\n";
	  MapIntStr::const_iterator tagiter = tags.end();
	  --tagiter;
	  untags.insert(std::make_pair(val, tagiter->first + 1));
     }
     return untags[val];
}

const std::string& Wordlist::unknownTag() const {
     return unknowntag;
}

const std::string& Wordlist::unknownLemma() const {
     return unknownlemma;
}

int Wordlist::getTag(const std::string& newword, const int optimal) {
     // Check part-of-speech. Use same if possible.
     Wordlist::const_iterator iter_first = findFirstWord(newword);
     Wordlist::const_iterator iter_last = findLastWord(newword);
     if (iter_first != end()) {
	  // Search for perfect match
	  for (Wordlist::const_iterator iter = iter_first;
	       iter != iter_last; ++iter) {
	       if (iter->tag() == optimal) {
		    return iter->tag();
	       }
	  }
	  // Perfect match not found. Just grab the first.
	  return iter_first->tag();
     }
     // Word not found. Unknown tag.
     return untag(unknownTag());
}

std::string Wordlist::preserveCase(const std::string& oldword, const std::string& newword) const {
     // Do nothing if no change is done
     if (oldword == newword) {
	  return newword;
     }
     std::string oldtest = oldword;
     std::string newtest = newword;
     // Check if all lower case
     Utils::lowercase(oldtest);
     if (oldword == oldtest) {
	  Utils::lowercase(newtest);
	  return newtest;
     }
     // Check if first letter is uppercase
     oldtest[0] = toupper(oldtest[0]);
     if (oldword == oldtest) {
	  Utils::lowercase(newtest);
	  newtest[0] = toupper(newtest[0]);
	  return newtest;
     }
     // Check if all upper case
     Utils::uppercase(oldtest);
     if (oldword == oldtest) {
	  Utils::uppercase(newtest);
	  return newtest;
     }
     // Else return lowercase
     Utils::lowercase(newtest);
     return newtest;
}

bool Wordlist::allUpper(const std::string& str) const {
     std::string test = str;
     Utils::uppercase(test);
     return (test == str);
}


